package de.smartrules.plugins.translation;

import java.io.FileInputStream;
import java.io.StringReader;
import java.io.StringWriter;
import java.text.MessageFormat;
import java.util.HashMap;
import java.util.Locale;
import java.util.PropertyResourceBundle;
import java.util.ResourceBundle;

import org.apache.commons.io.IOUtils;
import org.apache.commons.lang3.LocaleUtils;
import org.apache.log4j.Logger;

import com.appiancorp.services.ServiceContext;
import com.appiancorp.suiteapi.common.exceptions.InvalidVersionException;
import com.appiancorp.suiteapi.common.exceptions.PrivilegeException;
import com.appiancorp.suiteapi.content.Content;
import com.appiancorp.suiteapi.content.ContentConstants;
import com.appiancorp.suiteapi.content.ContentFilter;
import com.appiancorp.suiteapi.content.ContentService;
import com.appiancorp.suiteapi.content.exceptions.InvalidContentException;
import com.appiancorp.suiteapi.content.exceptions.InvalidTypeMaskException;
import com.appiancorp.suiteapi.expression.annotations.Function;
import com.appiancorp.suiteapi.expression.annotations.Parameter;
import com.appiancorp.suiteapi.knowledge.Document;
import com.appiancorp.suiteapi.personalization.UserDataType;
import com.appiancorp.suiteapi.personalization.UserProfileService;
import com.appiancorp.suiteapi.portal.AdministrationService;

@com.appiancorp.suiteapi.expression.annotations.AppianScriptingFunctionsCategory
public class Translate {
	private static final String LOG_NAME = Translate.class.getName();
	private static final Logger LOG = Logger.getLogger(LOG_NAME);

	private static HashMap<String, AppianDocumentBundle> cache = new HashMap<String, AppianDocumentBundle>();
	private static long last_read = System.currentTimeMillis();
	private static final long min_cache_time = 30*1000;  //never look at the db more than once every 30 seconds.  
														 // this is for performance on potentially large volumes of users.
	private static final String TRANSLATIONS_FOLDER_NAME = "Translations";
	private static Long translationsFolderId;

	@Function
	public String translateforuser(
		ServiceContext sc,
		ContentService cs,
		AdministrationService as,
		UserProfileService ups,
		@Parameter @UserDataType String context,
		@Parameter String bundle,
		@Parameter String key,
		@Parameter(required = false) String[] parameters
	) {
		if (context == null) {
			context = sc.getName();
		}
		Locale locale = ups.getUser(context).getLocale();
		/*if (locale == null) {
			//AdministrationService as = ServiceLocator.getPortalAdministrationService(sc);
			locale = as.getSiteLocaleSettings().getPrimaryLocale();
		}*/

		return getTranslation(cs, locale, bundle, key, parameters);
	}

	@Function
	public String translateforlocale(
		ContentService cs,
		@Parameter String locale,
		@Parameter String bundle,
		@Parameter String key,
		@Parameter(required = false) String[] parameters
	) {
		Locale l = LocaleUtils.toLocale(locale);

		return getTranslation(cs, l, bundle, key, parameters);
	}

	private String getTranslation(
		ContentService cs,
		Locale locale,
		String bundle,
		String key,
		String[] parameters
	) {
		ResourceBundle rb;
		try {
			rb = getBundle(cs, bundle, locale);
			String pattern = rb.getString(key);
			if (parameters != null && parameters.length > 0) {
				return MessageFormat.format(pattern, (Object[]) parameters);
			} else {
				return pattern;
			}
		} catch (Exception e) {
			LOG.error(e, e);
			return "???" + key + "???";
		}
	}

	private String getDocumentContent(Document aff) {
		try {
			String documentPath = aff.getInternalFilename();
			FileInputStream fis = new FileInputStream(documentPath);
			StringWriter sw = new StringWriter();
			IOUtils.copy(fis, sw);
			String form = sw.getBuffer().toString();
			return form;
		} catch (Exception e) {
			LOG.error(e, e);
		}
		return null;
	}

	private ResourceBundle getBundle(ContentService cs, String name, Locale locale) throws Exception {
		String tokens[] = name.split("\\.");
		if (tokens.length == 0) tokens = new String[] { name };

		AppianDocumentBundle bundle = cache.get(locale + name);
		if (bundle != null) {
			if (min_cache_time > (System.currentTimeMillis() - last_read)) {
				LOG.info("using cached version, current time to expire is "+(System.currentTimeMillis()-last_read)+"ms");
				return bundle.getBundle();
			}
			System.out.println("found a bundle for " + locale + name + ", checking version " + bundle.getDocVersion() + " ");
			Content lastVersion = cs.getVersion(bundle.getDocId(), ContentConstants.VERSION_CURRENT);
			last_read = System.currentTimeMillis();
			System.out.println(lastVersion.getId() + " " + lastVersion.getNumberOfVersions());
			if (lastVersion.getNumberOfVersions().equals(bundle.getDocVersion())) {
				LOG.info("using cached version");
				return bundle.getBundle();
			}
		}

		Long id = getTranslationsFolderId(cs);
		for (int i = 0; i < tokens.length; i++) {
			if (i < tokens.length - 1) {
				ContentFilter f = new ContentFilter(Content.TYPE_FOLDER);
				f.setName(tokens[i]);
				Content[] children = cs.getChildren(id, f,
						ContentConstants.GC_MOD_NORMAL);
				if (children.length == 0) {
					LOG.error("Cannot find folder " + tokens[i] + " int bundle key" + name);
				}
				id = children[0].getId();
			} else {
				// try just the whole locale
				ContentFilter f = new ContentFilter(Content.TYPE_DOCUMENT);
				f.setName(tokens[i] + "_" + locale.toString());
				f.setExtension(new String[] { "properties" });
				Content[] children = cs.getChildren(id, f,
						ContentConstants.GC_MOD_NORMAL);
				if (children.length == 0) {
					// try just the country code
					f.setName(tokens[i] + "_" + locale.getLanguage());
					children = cs.getChildren(id, f,
							ContentConstants.GC_MOD_NORMAL);
					if (children.length == 0) {
						// try the default
						f.setName(tokens[i]);
						children = cs.getChildren(id, f,
								ContentConstants.GC_MOD_NORMAL);
						if (children.length == 0) {
							LOG.error("Cannot find file for " + name + " and locale " + locale.toString());
							return null;
						}
					}
				}
				id = children[0].getId();
			}
		}

		Document aff = cs.download(id, ContentConstants.VERSION_CURRENT, true)[0];
		String s = getDocumentContent(aff);
		StringReader reader = new StringReader(s);

		ResourceBundle b = new PropertyResourceBundle(reader);
		bundle = new AppianDocumentBundle(id, aff.getNumberOfVersions(), b);
		cache.put(locale + name, bundle);
		return b;
	}

	private Long getTranslationsFolderId(ContentService cs) throws PrivilegeException, InvalidContentException, InvalidVersionException, InvalidTypeMaskException {
		if (translationsFolderId == null) {
			Content[] children = null;
			ContentFilter cf = null;
			
			Long communityRoot = cs.getIdByUuid(ContentConstants.UUID_COMMUNITY_ROOT);

			// Search for the translations KC
			cf = new ContentFilter(Content.TYPE_COMMUNITY_KC);
			cf.setName(TRANSLATIONS_FOLDER_NAME);
			children = cs.getChildren(communityRoot, cf, ContentConstants.GC_MOD_NORMAL);
			if(children.length > 0) {
				Long translationsKC = children[0].getId();

				// Search for the translations folder
				cf = new ContentFilter(Content.TYPE_FOLDER);
				cf.setName(TRANSLATIONS_FOLDER_NAME);
				children = cs.getChildren(translationsKC, cf, ContentConstants.GC_MOD_NORMAL);

				if(children.length > 0) {
					translationsFolderId = children[0].getId();
				} else {
					LOG.error("Folder for translations not found. Create a folder named \"Translations\" inside Knowledge Center \"Translations\"");
				}
			} else {
				LOG.error("Knowledge Center for translations not found. Create a KC named \"Translations\"");
			}
		}

		return translationsFolderId;
	}
}
